#!/usr/bin/env python3

################################################################################
# Copyright (C) 2016-2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import subprocess
import re


class Interface:
    def __init__(self, name):
        self.name = name
        self.addrs = {}
        self.flags = []

    def __repr__(self):
        return self.name


def get_interfaces():
    def _parse_flags(line, iface):
        pass

    def _parse_line(line, iface):
        line = line.strip()
        if line.startswith('inet'):
            s = re.sub(r'\s+', ' ', line).split(' ')
            i = 0
            while i < len(s) - 1:
                key = s[i]
                value = s[i + 1]
                if value.startswith('0x'):
                    try:
                        v = ("00000000" +value[2:])[-8:]
                        value = ".".join([str(int(v[i:i + 2], 16)) for i in range(0, len(v), 2)])
                    except Exception as ignore:
                        pass
                iface.addrs[key] = value
                i += 2

    def _parse_lines(lines, ifaces):
        current_iface = None

        for line in lines:
            if len(line) > 0 and not line[0].isspace():
                iface_split = line.split(":")
                name = iface_split[0]

                current_iface = Interface(name)
                ifaces[name] = current_iface
                if len(iface_split) > 0:
                    _parse_flags(iface_split[1], current_iface)
            else:
                _parse_line(line, current_iface)

    ifaces = {}

    try:
        process = subprocess.Popen('ifconfig', shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        retval = process.wait()

        lines = process.stdout.readlines()
        lines = [line.decode('ascii') for line in lines]
        _parse_lines(lines, ifaces)

    except Exception as e:
        pass

    return ifaces


if __name__ == "__main__":
    ifaces = get_interfaces()

    for k, v in ifaces.items():
        if 'inet' in v.addresses:
            print(f"{k}: {v.addresses}")



################################################################################
# Copyright (C) 2016-2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import os
import socket
import fcntl
import struct
import threading
import subprocess
import traceback

from discovery.interfaces import get_interfaces


class Discovery:
    def __init__(self, typ):
        self._type = typ
        self._debug = False
        self._receiving_socket = None
        self._thread = None
        self._stop = False
        self._services = {}
        self._callbacks = {}

    def register_service(self, name, value):
        self._services[name] = value

    def deregister_service(self, name):
        del self._services[name]

    def register_dynamic_service(self, name, callback):
        self._callbacks[name] = callback

    def deregiser_dynamic_service(self, name):
        del self._callbacks[name]

    @staticmethod
    def _get_ip_address_from_interface(ifname):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(fcntl.ioctl(s.fileno(), 0x8915, bytearray(struct.pack('256s', bytes(ifname[:15], 'utf-8'))))[20:24])

    @staticmethod
    def _get_hostname():
        hostname = "UNKNOWN"
        if os.path.exists("/etc/hostname"):
            with open("/etc/hostname", "rt") as textFile:
                hostname = textFile.read()
        else:
            # noinspection PyBroadException
            try:
                hostname = subprocess.Popen('hostname', shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) \
                    .stdout.readlines()[0].decode('ascii')
            except Exception:
                pass

        hostname = hostname.split(".")[0].replace('\n', '')
        if hostname.startswith("gcc-"):
            hostname = hostname[4].upper() + hostname[5:]
            hostname = hostname.replace("-", "")

        return hostname

    @staticmethod
    def _get_ip_address(client_ip):
        client_ip = client_ip.split('.')

        def _calc_weight(ip1):
            j = 0
            while j < len(ip1) and j < len(client_ip) and ip1[j] == client_ip[j]:
                j += 1
            return j

        all_ips = [iface.addrs['inet'].split('.') for iface in get_interfaces().values() if 'inet' in iface.addrs and 'broadcast' in iface.addrs]
        ip_weights = {".".join(ip): _calc_weight(ip) for ip in all_ips}

        return max(ip_weights, key=lambda k: ip_weights[k])

    def _prepare_response(self, client_ip):
        response_map = {
            "IP": self._get_ip_address(client_ip),
            "NAME": self._get_hostname(),
            "TYPE": self._type}

        for service_name in self._services:
            response_map[service_name] = self._services[service_name]

        for callback_name in self._callbacks:
            callback = self._callbacks[callback_name]
            callback(response_map)

        return "A#" + ";".join([k + "=" + v for k, v in response_map.items()])

    def _receive(self):
        self._receiving_socket.settimeout(10)
        if self._debug:
            print("    Started receive thread...")
        while not self._stop:
            # noinspection PyBroadException
            try:
                data, addr = self._receiving_socket.recvfrom(1024)
                request = str(data, 'utf-8')

                if self._debug:
                    print("Received: " + str(request))

                if request.startswith("Q#"):
                    p = {kv[0].strip(): kv[1].strip() for kv in [line.split("=") for line in [entry.replace('\n', '') for entry in request[2:].split(';')]]}

                    return_ip = p["IP"]
                    # noinspection PyBroadException
                    try:
                        return_port = int(p["PORT"])
                    except Exception:
                        return_port = 0xd15c

                    response = self._prepare_response(return_ip)

                    self._send(return_ip, return_port, response)
            # except Exception as ex:
            #     print("ERROR: " + str(ex) + "\n" + ''.join(traceback.format_tb(ex.__traceback__)))
            except Exception:
                pass

    def _send(self, ip, port, packet):
        if self._debug:
            print("Debug: sending packet to " + str(ip) + ":" + str(port) + " " + packet)

        sending_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sending_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sending_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sending_socket.setblocking(False)

        sending_socket.sendto(bytes(packet, 'utf-8'), (ip, port))

    def start(self):
        self._stop = False
        self._receiving_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._receiving_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._receiving_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self._receiving_socket.bind(('', 0xd15c))

        self._thread = threading.Thread(target=self._receive, args=())
        self._thread.daemon = True
        self._thread.start()

    def stop(self):
        self._stop = True



#!/usr/bin/env python3

################################################################################
# Copyright (C) 2016-2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import socket
import threading
import time
# import netifaces
from discovery.interfaces import get_interfaces


DEFAULT_DISCOVERY_TIMEOUT = 5


class Discover:
    def __init__(self, discovery_timeout, debug=False):
        self._debug = debug
        self._discovery_timeout = discovery_timeout
        self._discovery_sockets = []
        self._responses = []

        ifaces = get_interfaces()

        if self._debug:
            print("  Discovered network adapters:")

        for iface_name in ifaces:
            iface = ifaces[iface_name]
            if 'inet' in iface.addrs and 'broadcast' in iface.addrs and iface.addrs['inet'] != '127.0.0.1':
                my_ip = iface.addrs['inet']
                broadcast_ip = iface.addrs['broadcast']

                sending_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sending_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sending_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                sending_socket.setblocking(False)
                sending_socket.settimeout(self._discovery_timeout)

                listening_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                listening_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                listening_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
                listening_socket.setblocking(False)
                listening_socket.settimeout(self._discovery_timeout)

                listening_socket.bind((my_ip, 0))

                self._discovery_sockets.append((my_ip, broadcast_ip, sending_socket, listening_socket))

                if self._debug:
                    print("    " + str(broadcast_ip))

    @staticmethod
    def _apply_netmask(ip, netmask, value):
        ip_a = ip.split('.')
        netmask_a = netmask.split('.')
        for i in range(0, 4):
            if netmask_a[i] != '255':
                ip_a[i] = value

        return ".".join(ip_a)

    def send_discovery_packet(self, packet):
        for i in range(0, len(self._discovery_sockets)):
            my_ip, broadcast_ip, sending_socket, listening_socket = self._discovery_sockets[i]

            updated_packet = packet + "IP=" + my_ip + ";PORT=" + str(listening_socket.getsockname()[1])
            sending_socket.sendto(bytes(updated_packet, 'utf-8'), (broadcast_ip, 0xd15c))
            if self._debug:
                print("  send discovery packet to " + str(broadcast_ip) + ":0xd15c, packet=" + str(updated_packet))

    def _process_response(self, response_to_processs, callback=None):
        parsed_response = {kv[0]: kv[1] for kv in [kv for kv in [kv.split('=') for kv in response_to_processs.split(';')] if len(kv) > 1]}
        parsed_response.update({kv[0]: None for kv in [kv for kv in [kv.split('=') for kv in response_to_processs.split(';')] if len(kv) == 1]})
        # parsed_response.update({"ERROR_" + str(i + 1): kv[0] for i, kv in enumerate([kv for kv in [kv.split('=') for kv in response_to_processs.split(';')] if len(kv) == 1])})

        self._responses.append(parsed_response)
        if callback is not None:
            callback(parsed_response)

    def _receive_discovery_packets(self, listening_socket, callback=None):
        start_time = time.time()
        if self._debug:
            print("  receiving at " + str(listening_socket.getsockname()))
        while time.time() - start_time < self._discovery_timeout:
            # noinspection PyBroadException
            try:
                data, addr = listening_socket.recvfrom(1024)
                p = str(data, 'utf-8')
                if self._debug:
                    print("  received " + p)

                if p.startswith("A#"):
                    self._process_response(p[2:], callback)
                # elif p.startswith("Q#"):
                #     print("Received self query:" + p)
            except Exception:
                pass

    def discover(self, callback=None):
        del self._responses[:]
        for my_ip, broadcast_ip, sending_socket, listening_socket in self._discovery_sockets:
            thread = threading.Thread(target=self._receive_discovery_packets, args=(listening_socket, callback))
            thread.daemon = False
            thread.start()

        time.sleep(1)
        self.send_discovery_packet("Q#")

        time.sleep(self._discovery_timeout + 0.2)

        return self._responses


if __name__ == "__main__":
    discover = Discover(10)
    response = discover.discover()
    print("")
    print("{0!s:<20} {1:<20} {2:<20} {3}".format("ip", "name", "type", "other"))
    for details in response:
        # print("{0!s:<20} {1:<20} {2:<20} {3}".format(details["IP"] + ":" + details["PORT"], details["NAME"], details["TYPE"], ";".join(errors)))
        # print("{0!s:<20} {1:<20} {2:<20} {3}".format(details["IP"] + ":" + details["PORT"] if "PORT" in details else "", details["NAME"], details["TYPE"], ""))
        print(str(details))
